﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2012 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;

namespace HeuristicLab.EvolutionaryTracking {
  public class GenealogyGraphNode : IGenealogyGraphNode {
    public string Id { get; private set; }
    public string Label { get; set; }
    public List<GenealogyGraphArc> InEdges { get; private set; }
    public List<GenealogyGraphArc> OutEdges { get; private set; }
    private object _content;
    public object Content {
      get { return _content; }
      set {
        if (value == null)
          throw new ArgumentException("Node content cannot be null.");
        _content = value;
      }
    }

    public GenealogyGraphNode() {
      Id = Guid.NewGuid().ToString();
    }

    public GenealogyGraphNode(GenealogyGraphNode node) {
      Id = Guid.NewGuid().ToString();
      Label = node.Label;
      Content = node.Content;
      InEdges = new List<GenealogyGraphArc>(node.InEdges);
      OutEdges = new List<GenealogyGraphArc>(node.OutEdges);
    }

    public int InDegree { get { return InEdges == null ? 0 : InEdges.Count; } }
    public int OutDegree { get { return OutEdges == null ? 0 : OutEdges.Count; } }
    public int Degree { get { return InDegree + OutDegree; } }

    /// <summary>
    /// Performs an upward-breath-traversal of the genealogy graph using the nodes InEdges
    /// </summary>
    /// <returns>All the ancestors of the current node</returns>
    public IEnumerable<IGenealogyGraphNode> Ancestors() {
      // for performance, we use a hashset for lookup and a list for iteration
      var nodes = new HashSet<IGenealogyGraphNode> { this };
      var list = new List<IGenealogyGraphNode> { this };
      int i = 0;
      while (i != list.Count) {
        if (list[i].InEdges != null) {
          foreach (var e in list[i].InEdges) {
            if (nodes.Contains(e.Target)) continue;
            nodes.Add(e.Target);
            list.Add(e.Target);
          }
        }
        ++i;
      }
      return list;
    }

    /// <summary>
    /// Performs a downward-breath-traversal of the genealogy graph using the nodes OutEdges
    /// </summary>
    /// <returns>All the descendants of the current node</returns>
    public IEnumerable<IGenealogyGraphNode> Descendants() {
      var nodes = new HashSet<IGenealogyGraphNode> { this };
      var list = new List<IGenealogyGraphNode> { this };
      int i = 0;
      while (i != list.Count) {
        if (list[i].OutEdges != null) {
          foreach (var e in list[i].OutEdges) {
            if (nodes.Contains(e.Target)) continue;
            nodes.Add(e.Target);
            list.Add(e.Target);
          }
        }
        ++i;
      }
      return list;
    }

    // this method can be used to add arcs towards targets that are not visible in the graph 
    // (they do not appear in the _nodes Dictionary). It is the programmers responsibility to 
    // enforce the correct and desired behavior.
    public void AddForwardArc(IGenealogyGraphNode target, object data = null) {
      var e = new GenealogyGraphArc { Target = target, Data = data };
      if (OutEdges == null) OutEdges = new List<GenealogyGraphArc> { e };
      else OutEdges.Add(e);
    }

    public void AddReverseArc(IGenealogyGraphNode target, object data = null) {
      var e = new GenealogyGraphArc { Target = target, Data = data };
      if (InEdges == null) InEdges = new List<GenealogyGraphArc> { e };
      else InEdges.Add(e);
    }
  }
}
